{
  Copyright (c) 2017-2022, Krzysztof Jacewicz <k.k.jacewicz@gmail.com>
  All rights reserved.

  This file is part of All Things Ubuntu (ATU) project package and
  is released under MIT License, please refer to the file COPYING.txt
  for the license text.

  More info: https://sourceforge.net/projects/all-things-ubuntu-library/

  SYNOPSIS:
    This unit will extract VS_VERSION_INFO data from a binary file. Without using
    Lazarus Resources. It will read binary and ASCII data embedded in a compiled
    executable, locate the VS_VERSION_INFO record, and read directly.

  COMPATIBILITY:
    2022-04-26:
      Only tested against VERSION_INFO compiled into executables by Lazarus IDE,
      not sure if it also works for other executables.
      Tested on Linux amd86_64 and armhf, should compile everywhere.
      Tested against Linux ELF executables, not tested on EXE files yet.
}
unit LazVersionInfo;
{$mode objfpc}{$H+}
interface
uses
  Classes, SysUtils;

const
  VS_KEYS       : array[0..10] of string
                = ('VS_VERSION_INFO' //a marker for the VS_VERSION_INFO structure compiled into the executable in a BTSTR format (every character word sized utf-16 encoded value )
                  ,'Comments'
                  ,'CompanyName'
                  ,'FileDescription'
                  ,'FileVersion'
                  ,'InternalName'
                  ,'LegalCopyright'
                  ,'LegalTrademarks'
                  ,'OriginalFilename'
                  ,'ProductName'
                  ,'ProductVersion');
  wTypeBinary     = 0;
  wTypeText       = 1;

type
  TChars = array of char;
  PChars = ^TChars;
  TVersionInfoItem
  = record
      KeyName          : string;
      StructLength     : word;
      ValueLength      : word;
      ValueType        : word;
      RawValue         : pchar;
      TextValue        : string;
    end;
  PVersionInfoItem = ^TVersionInfoItem;

  TBinaryVersion
  = record
      major   : word;
      minor   : word;
      release : word;
      build   : word;
    end;

  TVersionInfo
  = record
      BinaryFile        : string;
      VersionInfoOffset : int64;
      Comments          : TVersionInfoItem;
      CompanyName       : TVersionInfoItem;
      FileDescription   : TVersionInfoItem;
      FileVersion       : TVersionInfoItem;
      FileVersionBin    : TBinaryVersion;
      InternalName      : TVersionInfoItem;
      LegalCopyright    : TVersionInfoItem;
      LegalTrademarks   : TVersionInfoItem;
      OriginalFilename  : TVersionInfoItem;
      ProductName       : TVersionInfoItem;
      ProductVersion    : TVersionInfoItem;
    end;

{ TEmbeddedVersionInfo }
  TEmbeddedVersionInfo
  = Class
      Protected
        FFilename       : string;
        FHasVersionInfo : Boolean;
        FKeys           : array of PVersionInfoItem;
        FVersionInfo    : TVersionInfo;
      Private
        Function GetComments: string;
        Function GetCompanyName: string;
        Function GetFileDescription: string;
        Function GetFileVersion: string;
        Function GetInternalName: string;
        Function GetLegalCopyright: string;
        Function GetLegalTrademarks: string;
        Function GetOriginalFilename: string;
        Function GetProductName: string;
        Function GetProductVersion: string;
        Function GetMajor: word;
        Function GetMinor: word;
        Function GetRelease: word;
        Function GetBuild: word;
        Function GetVersion(Index: byte): word;
        Function GetCount: integer;
        Function GetKey(const Index: integer): string;
        Function GetValue(const Key: string): string;
        Function GetValueByIndex(const Index: integer): string;
        Function GetVersionInfoOffset: int64;
      Public
        Constructor Create(AFilename: string);
        Destructor Destroy; override;
      //METHODS:
        Function LoadFromFile(const AFilename: string): Boolean;
      //PROPERTIES:
        Property Version[Index: byte]: word read GetVersion;
        Property VersionMajorNumber: word read GetMajor;
        Property VersionMinorNumber: word read GetMinor;
        Property VersionReleaseNumber: word read GetRelease;
        Property VersionBuildNumber: word read GetBuild;
        Property VersionInfoOffset: int64 read GetVersionInfoOffset;
        Property KeyCount: integer read GetCount;
        Property Keys[const Index: integer]: string read GetKey; default;
        Property Values[const Key: string]: string read GetValue;
        Property ValueByIndex[const Index: integer]: string read GetValueByIndex;
        Property HasVersionInfo: Boolean read FHasVersionInfo;
        Property Comments: string read GetComments;
        Property CompanyName: string read GetCompanyName;
        Property FileDescription: string read GetFileDescription;
        Property FileVersion: string read GetFileVersion;
        Property InternalName: string read GetInternalName;
        Property LegalCopyright: string read GetLegalCopyright;
        Property LegalTrademarks: string read GetLegalTrademarks;
        Property OriginalFilename: string read GetOriginalFilename;
        Property ProductName: string read GetProductName;
        Property ProductVersion: string read GetProductVersion;
    End;

{ API }
  Function TryReadVersionInfoFromFile(AFilename: string; out AResult: TVersionInfo): Boolean;

implementation
{ API }
  Function ReadMarkerFromBuf(const BUF: array of byte; p: int64; const Marker: string; const SwapBytes: Boolean): Boolean;
    var
      i,ln : int64;
    begin
      Result := False;
      if(Marker='')then exit;
      ln := Length(Marker);
      if(p<0)or(p>=Length(buf)-ln)then exit;
      for i:=1 to ln do
        begin
          if(SwapBytes)
            then begin
                   if(BUF[(p+0)]<>0)then exit;
                   if(BUF[(p+1)]<>ord(Marker[i]))then exit;
                 end
            else begin
                   if(BUF[(p+1)]<>0)then exit;
                   if(BUF[(p+0)]<>ord(Marker[i]))then exit;
                 end;
          inc(p,2);
        end;
      Result := True;
    end;

  Function FindMarker(const buf: array of byte; const AMarker: string; const SwapBytes: Boolean): int64;
    var
      i : int64;
    begin
      Result := -1;
      if(Length(AMarker)<1)then exit;
      if(Length(buf)-1-Length(AMarker)<=0)then exit;
      for i:=0 to (Length(buf)-1-Length(AMarker))do
        begin
          if not ReadMarkerFromBuf(buf, i, AMarker, SwapBytes)then continue;
          Result := i;
          exit;
        end;
    end;

  Function FindMarker(const FS: TStream; const AMarker: string; const SwapBytes: Boolean): int64;
    const
      BUFSIZE = 4096;
      { The reason we are using the buffer is because changing position of
      * read cursor in a file, and repeatedly reading small pieces of data
      * this way is very slow.
      * So instad, we load bigger chunks from a file into memory and do the
      * search there in a fast way. }
    var
      buf : array of byte;
      so  : int64; //scan offset (how many bytes have already been scanned before)
      br  : int64; //bytes read
      p   : int64;
    begin
      Result := -1;
      if not Assigned(FS)then exit;
      if(AMarker='')then exit;
      FS.Position := 0;
      so          := 0;
      SetLength(buf, 0);
      try
        repeat
          p  := FS.Size - FS.Position; if(p>BUFSIZE)then p := BUFSIZE;
          if(Length(buf)<>p)then SetLength(buf, p);
          try
            br := FS.Read(buf[0], Length(buf));
          except
            writeln(stderr, 'Error: buffer reading exception raised.');
          end;
          if(br<(Length(AMarker)*3))then exit;
          if(br<>Length(buf))then SetLength(buf, br);
          p := FindMarker(buf, AMarker, SwapBytes);
          if(p>=0)then
            begin
              Result := so+p;
              exit;
            end;
          p := FS.Position-(Length(AMarker)*3); if(p<0)then p := 0;
          FS.Position := p;
          so := FS.Position;
        until(FS.Position>=FS.Size);
      finally
        SetLength(buf,0);
      end;
    end;

  Function BTSTRtoStr(const buf: TChars; wOffset: word; const ASwapBytes: Boolean): UnicodeString;
   var
     ln,i : integer;
     sz   : PWideChar;
   begin
     Result := '';
     ln := Length(buf);
     sz := @buf[0];
     if(wOffset>0)then
       begin
         inc(sz,wOffset);
         dec(ln, wOffset*SizeOf(word)); //because wOffset is in words, and ln is in bytes
       end;
     If(sz^=#0)then inc(sz);
     ln := ln div SizeOf(word); //converting bytes to words
     if(ln<1)then exit;
   //to do: --------------------------------------------------------------------
   {        Need to test if this is neccessary. Even if byte order is different
            due to a different endianness, maybe BTSTR is already uniform      }

     if(ASwapBytes)
       then for i:=0 to ln-1 do Swap(word(sz[i]));
   //to do: --------------------------------------------------------------------
     Result := trim(Copy(sz, 0, ln));
   end;


  Function TryReadVersionInfoAttributeEx(const FS: TStream; Offset: int64; const AKey: string; var AValue: TVersionInfoItem; const ASwapBytes: Boolean): Boolean;
   var
     p  : int64;
     w  : array[0..2]of word;
     ln : integer;
     L  : integer;
     buf: TChars;
   begin
      Result := False;
     if not Assigned(FS)then exit;
     if(Offset<0)then Offset:=0;
     ln := Length(AKey);
     if(ln<1)then exit;
     if(Offset>=FS.Size-Length(AKey))then exit;
     p := -1;
     try
       AValue.KeyName   := AKey;
       while(p<0)do
         begin
           FS.Position := Offset;
           p := FindMarker(FS, AKey, ASwapBytes);
           if(p<0)then exit;
           FS.Position := p-(SizeOf(word)*Length(w));
           FS.Read(w[0], Length(w)*SizeOf(word));
           if(ASwapBytes)
             then for L:=Low(w)to High(w) do w[L] := Swap(w[L]); { endiannes conversion }
           if(w[2]<>$0001)or(w[1]<2)then begin Offset := p+1; continue; end;
           L := w[0]-(SizeOf(word)*Length(w));
           SetLength(buf, L); FillChar(buf[0], Length(buf), 0);
           //FS.Position := p;//FS.Position + (2*(ln+1));
           FS.Read(buf[0], L);
           if(buf[L-1]<>#0)and(buf[L-2]<>#0)then begin Offset := p+1; continue; end;
           AValue.StructLength := w[0];
           AValue.ValueLength  := w[1];
           AValue.ValueType    := w[2];
           AValue.TextValue    := UTF8Encode(BTSTRtoStr(buf, (Length(AKey)+1), ASwapBytes)); //+1 is for null termination of the string
           AValue.RawValue     := nil;
           Result              := True;
         end;
     finally
       SetLength(buf,0);
     end;
   end;

  Function TryReadVersionInfoAttribute(const FS: TStream; Offset: int64; const AKey: string; var AValue: TVersionInfoItem): Boolean;
    begin
      Result := TryReadVersionInfoAttributeEx(FS, Offset, AKey, AValue, false);
      if(Result)then exit;
      Result := TryReadVersionInfoAttributeEx(FS, Offset, AKey, AValue, true);
    end;

  Function TryReadVersionInfoAttribute(const FS: TStream; Offset: int64; const AKey: string; out AValue: string): Boolean;
    var
      VII : TVersionInfoItem;
    begin
      Result := TryReadVersionInfoAttribute(FS, Offset, AKey, VII);
      if Result then AValue := VII.TextValue;
    end;

  Function ReadVersionInfoAttributeDef(const FS: TStream; Offset: int64; const AKey: string; const ADefault: string = ''): string;
   begin
     if not TryReadVersionInfoAttribute(FS, Offset, AKey, Result)
       then Result := ADefault;
   end;

  Function LoadBinaryersion(VerStr: string; var BinVer: TBinaryVersion): Boolean;
    var
      p : integer;
      v : array[1..4]of integer;
      s : string;
    begin
      Result         := False;
      BinVer.major   := 0;
      BinVer.minor   := 0;
      BinVer.release := 0;
      BinVer.build   := 0;
      VerStr := trim(VerStr);
    //major:
      p := Pos('.', VerStr); if(p<2)then exit;
      s := Copy(VerStr, 1, p-1); Delete(VerStr,2,p);
      if not TryStrToInt(s, v[1]) then exit;
      if(v[1]<0)then exit;
    //minor:
      p := Pos('.', VerStr); if(p<2)then exit;
      s := Copy(VerStr, 1, p-1); Delete(VerStr,2,p);
      if not TryStrToInt(s, v[2]) then exit;
      if(v[2]<0)then exit;
    //release:
      p := Pos('.', VerStr); if(p<2)then exit;
      s := Copy(VerStr, 1, p-1); Delete(VerStr,2,p);
      if not TryStrToInt(s, v[3]) then exit;
      if(v[4]<0)then exit;
    //build:
      if not TryStrToInt(s, v[4]) then exit;
      if(v[4]<0)then exit;
    //Success:
      BinVer.major   := v[1];
      BinVer.minor   := v[2];
      BinVer.release := v[3];
      BinVer.build   := v[4];
      Result         := True;
    end;

  Function StreamCopyFrom(const AFrom,AInto: TStream; const ACount: int64 = 4096): int64;
    var
      buf : array of byte;
      len : int64;
    begin
      Result := 0;
      try
        if not Assigned(AFrom)then exit;
        if not Assigned(AInto)then exit;
        if(ACount<1)then exit;
        SetLength(buf, ACount);
        len := AFrom.Read(buf[0], ACount);
        AInto.write(buf[0], len);
      finally
        SetLength(buf,0);
      end;
    end;

  Function TryReadVersionInfoFromFile(AFilename: string; out AResult: TVersionInfo): Boolean;
    var
      FS  : TFileStream;
      VIO : int64; //VersionInfoOffset
      MS  : TMemoryStream;
      SB  : Boolean;
    begin
      Result := False; SB := False; //don't swap bytes
      if(AFileName='')then exit;
      if not FileExists(AFilename)then exit;
      FS := TFileStream.Create(AFilename, fmOpenRead);
      try
        try
          VIO := FindMarker(FS, VS_KEYS[0], SB);
          if(VIO<0)then VIO := FindMarker(FS, VS_KEYS[0], not SB);
          if(VIO<0)then exit;
        //Cache data from file for speed:
          MS := TMemoryStream.Create;
          try
            FS.Position := VIO;
            StreamCopyFrom(FS, MS);
            MS.Position := 0;
            AResult.VersionInfoOffset := VIO;
            Result := TryReadVersionInfoAttribute(MS, 0, VS_KEYS[ 1], AResult.Comments) or Result;
            Result := TryReadVersionInfoAttribute(MS, 0, VS_KEYS[ 2], AResult.CompanyName) or Result;
            Result := TryReadVersionInfoAttribute(MS, 0, VS_KEYS[ 3], AResult.FileDescription) or Result;
            Result := TryReadVersionInfoAttribute(MS, 0, VS_KEYS[ 4], AResult.FileVersion) or Result;
            Result := TryReadVersionInfoAttribute(MS, 0, VS_KEYS[ 5], AResult.InternalName) or Result;
            Result := TryReadVersionInfoAttribute(MS, 0, VS_KEYS[ 6], AResult.LegalCopyright) or Result;
            Result := TryReadVersionInfoAttribute(MS, 0, VS_KEYS[ 7], AResult.LegalTrademarks) or Result;
            Result := TryReadVersionInfoAttribute(MS, 0, VS_KEYS[ 8], AResult.OriginalFilename) or Result;
            Result := TryReadVersionInfoAttribute(MS, 0, VS_KEYS[ 9], AResult.ProductName) or Result;
            Result := TryReadVersionInfoAttribute(MS, 0, VS_KEYS[10], AResult.ProductVersion) or Result;
            LoadBinaryersion(AResult.FileVersion.TextValue, AResult.FileVersionBin);
          finally
            FreeAndNil(MS);
          end;
        finally
          FreeAndNil(FS);
        end;
      Except
        exit;
      end;
    end;

{ TEmbeddedVersionInfo }
  Constructor TEmbeddedVersionInfo.Create(AFilename: string);
    begin
      inherited Create;
      SetLength(FKeys, 10);
      FKeys[0] := @FVersionInfo.Comments;
      FKeys[1] := @FVersionInfo.CompanyName;
      FKeys[2] := @FVersionInfo.FileDescription;
      FKeys[3] := @FVersionInfo.FileVersion;
      FKeys[4] := @FVersionInfo.InternalName;
      FKeys[5] := @FVersionInfo.LegalCopyright;
      FKeys[6] := @FVersionInfo.LegalTrademarks;
      FKeys[7] := @FVersionInfo.OriginalFilename;
      FKeys[8] := @FVersionInfo.ProductName;
      FKeys[9] := @FVersionInfo.ProductVersion;
      FHasVersionInfo := self.LoadFromFile(AFilename);
    end;

  Destructor TEmbeddedVersionInfo.Destroy;
    begin
      SetLength(FKeys, 0);
      inherited;
    end;

  Function TEmbeddedVersionInfo.LoadFromFile(const AFilename: string): Boolean;
    begin
      Result := False;
      try
        FFilename := trim(AFilename);
        if(FFilename='')or(not FileExists(FFilename))then exit;
        Result    := TryReadVersionInfoFromFile(FFilename, FVersionInfo);
      finally
        FHasVersionInfo := Result;
      end;
    end;

  Function TEmbeddedVersionInfo.GetComments: string;
    begin
      if(HasVersionInfo)
        then Result := FVersionInfo.Comments.TextValue
        else Result := '';
    end;

  Function TEmbeddedVersionInfo.GetCompanyName: string;
    begin
      if(HasVersionInfo)
        then Result := FVersionInfo.CompanyName.TextValue
        else Result := '';
    end;

  Function TEmbeddedVersionInfo.GetFileDescription: string;
    begin
      if(HasVersionInfo)
        then Result := FVersionInfo.FileDescription.TextValue
        else Result := '';
    end;

  Function TEmbeddedVersionInfo.GetFileVersion: string;
    begin
      if(HasVersionInfo)
        then Result := FVersionInfo.FileVersion.TextValue
        else Result := '';
    end;

  Function TEmbeddedVersionInfo.GetInternalName: string;
    begin
      if(HasVersionInfo)
        then Result := FVersionInfo.InternalName.TextValue
        else Result := '';
    end;

  Function TEmbeddedVersionInfo.GetLegalCopyright: string;
    begin
      if(HasVersionInfo)
        then Result := FVersionInfo.LegalCopyright.TextValue
        else Result := '';
    end;

  Function TEmbeddedVersionInfo.GetLegalTrademarks: string;
    begin
      if(HasVersionInfo)
        then Result := FVersionInfo.LegalTrademarks.TextValue
        else Result := '';
    end;

  Function TEmbeddedVersionInfo.GetOriginalFilename: string;
    begin
      if(HasVersionInfo)
        then Result := FVersionInfo.OriginalFilename.TextValue
        else Result := '';
    end;

  Function TEmbeddedVersionInfo.GetProductName: string;
    begin
      if(HasVersionInfo)
        then Result := FVersionInfo.ProductName.TextValue
        else Result := '';
    end;

  Function TEmbeddedVersionInfo.GetProductVersion: string;
    begin
      if(HasVersionInfo)
        then Result := FVersionInfo.ProductVersion.TextValue
        else Result := '';
    end;

  Function TEmbeddedVersionInfo.GetMajor: word;
    begin
      Result := Version[0];
    end;

  Function TEmbeddedVersionInfo.GetMinor: word;
    begin
      Result := Version[1];
    end;

  Function TEmbeddedVersionInfo.GetRelease: word;
    begin
      Result := Version[2];
    end;

  Function TEmbeddedVersionInfo.GetBuild: word;
    begin
      Result := Version[3];
    end;

  Function TEmbeddedVersionInfo.GetVersion(Index: byte): word;
    begin
      Result := 0;
      if(not HasVersionInfo)then exit;
      case Index of
        0 : Result := FVersionInfo.FileVersionBin.Major;
        1 : Result := FVersionInfo.FileVersionBin.minor;
        2 : Result := FVersionInfo.FileVersionBin.release;
        3 : Result := FVersionInfo.FileVersionBin.build;
        else Result := 0;
      end;
    end;

  Function TEmbeddedVersionInfo.GetCount: integer;
    begin
      Result := Length(FKeys);
    end;

  Function TEmbeddedVersionInfo.GetKey(const Index: integer): string;
    begin
      if(Index<Low(FKeys))or(Index>High(FKeys))or(FKeys[Index]=nil)
        then Result := ''
        else Result := FKeys[Index]^.KeyName;
    end;

  Function TEmbeddedVersionInfo.GetValue(const Key: string): string;
    var
      i : integer;
    begin
      Result := '';
      for i:=0 to Length(FKEYS)-1 do
        begin
          if(FKeys[i]=nil)then continue;
          if(FKeys[i]^.KeyName<>Key)then continue;
          Result := FKeys[i]^.TextValue;
          exit;
        end;
    end;

  Function TEmbeddedVersionInfo.GetValueByIndex(const Index: integer): string;
    begin
      if(Index<Low(FKeys))or(Index>High(FKeys))or(FKeys[Index]=nil)
        then Result := ''
        else Result := FKeys[Index]^.TextValue;
    end;

  Function TEmbeddedVersionInfo.GetVersionInfoOffset: int64;
    begin
      if(HasVersionInfo)
        then Result := FVersionInfo.VersionInfoOffset
        else Result := -1;
    end;

end.

